/*
 * Main author(s) : Jean-Noël Monette
 * 
 * Copyright (c) 2013, Jean-Noël Monette
 * All rights reserved.

 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package convexPairs.inoscar.modeling

import oscar.cp.core.CPVarInt
import oscar.cp.core.Constraint
import convexPairs.inoscar.examples._
import java.util.LinkedList
import convexPairs.inoscar.GenericPropagator
import convexPairs.inoscar.GenericPropagatorWithIndices

trait Constraints {
	/**
     * Constraint enforcing n * sum,,i,, |x[i]-s/n| <= nd and sum,,i,, x[i] = s <br> 
     * Note that this constraint is very similar to spread.
     * @param x
     * @param s
     * @param nd
     */
    def deviation2(x : Iterable[CPVarInt], s: Int, nd: CPVarInt): Constraint = {
      new GenericPropagator(x.toArray, new DeviationFunction(s,x.toArray),nd,0,null,s,null,s, "Deviation Generic", false);
    }
    def deviation3(x : Iterable[CPVarInt], s: Int, nd: CPVarInt): Constraint = {
      new GenericPropagatorWithIndices(x.toArray, new DeviationFunction(s,x.toArray),nd,0,null,s,null,s, "Deviation Generic", false);
    }
    

    /**
     * Constraint enforcing sum,,i,, x[i]^2^ <= s2 and sum,,i,, x[i] = s <br> 
     * Note that this constraint is very similar to deviation.
     * @param x
     * @param s
     * @param s2
     */
    def spread2(x : Iterable[CPVarInt], s: Int, s2 : CPVarInt): Constraint = {
    	val n = x.size;
    	val sp = s2*(n*n)-(n*s*s);
    	new GenericPropagator(x.toArray, new SpreadFunction(s,x.toArray),sp,0,null,s,null,s, "Spread Generic", false);
    }
	def spread3(x : Iterable[CPVarInt], s: Int, s2 : CPVarInt): Constraint = {
    	val n = x.size;
    	val sp = s2*(n*n)-(n*s*s);
    	new GenericPropagatorWithIndices(x.toArray, new SpreadFunction(s,x.toArray),sp,0,null,s,null,s, "Spread Generic", false);
    }

    /**
     * Constraint enforcing sum_i a[i]*x[i] = s1 and sum_i x[i] = s2 
     */
    def linearAndSum(x : Iterable[CPVarInt], s1: CPVarInt, s2: CPVarInt, a: Iterable[Int], sort: Boolean = true): LinkedList[Constraint] = {
    	val cons = new LinkedList[Constraint];
    	cons.add(new GenericPropagator(x.toArray, new LinearAndSumFunction(a.toArray, x.toArray), s1, 0, s2, 0, s2, 0, "LinearAndSum Part1", sort));
    	cons.add(new GenericPropagator(x.toArray, new LinearAndSumFunction(a.map(v => -v).toArray, x.toArray), s1.opposite(), 0, s2, 0, s2, 0, "LinearAndSum Part1", sort));
    	cons;
    }
    
    
    def linear(x: Iterable[CPVarInt], s:CPVarInt, a:Iterable[Int], sort: Boolean = true): LinkedList[Constraint] = {
    	val cons = new LinkedList[Constraint];
    	cons.add(new GenericPropagator(x.toArray, new LinearAloneFunction(a.toArray, x.toArray), s, 0, null, 0, null, 0, "Linear Part1", sort));
    	cons.add(new GenericPropagator(x.toArray, new LinearAloneFunction(a.map(v => -v).toArray, x.toArray), s.opposite, 0, null, 0, null, 0, "Linear Part2", sort));
    	cons
    }
    
    def sumMy(x: Iterable[CPVarInt], s: CPVarInt): LinkedList[Constraint] = {
    	val cons = new LinkedList[Constraint];
    	cons.add(new GenericPropagator(x.toArray, new SumAloneC1Function(x.toArray), s, 0, null, 0, null, 0, "Sum Part1", false));
    	cons.add(new GenericPropagator(x.toArray, new SumAloneC1Function(x.map(_.opposite()).toArray), s.opposite, 0, null, 0, null, 0, "Sum Part2", false));
    	cons
    }
    def sumMy2(x: Iterable[CPVarInt], s: CPVarInt): Constraint = {
    	new GenericPropagator(x.toArray, new SumAloneC2Function(x.toArray), null, 0, s, 0, s, 0, "Sum", false)
    }
    
    def linearAndCount(x : Iterable[CPVarInt], s1: CPVarInt, s2: CPVarInt, a: Iterable[Int], V:Int, sort: Boolean = true): LinkedList[Constraint] = {
    	val cons = new LinkedList[Constraint];
    	cons.add(new GenericPropagator(x.toArray, new LinearAndCountFunction(a.toArray, V, x.toArray), s1, 0, s2, 0, s2, 0, "LinearAndCount Part1", sort));
    	cons.add(new GenericPropagator(x.toArray, new LinearAndCountFunction(a.map(v => -v).toArray, V, x.toArray), s1.opposite(), 0, s2, 0, s2, 0, "LinearAndCount Part1", sort));
    	cons;
    }
}
